#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

Display *dpy;
Window root;

static int * buttons[256];
static int debug =0;

static void 
parse( char * string, int button, int offset)
{
   char * d;
   int mod = 0;
   KeySym k; 
    
   while (( d = strchr( string, '+'))) {
      *d = 0;
      if ( strcmp( string, "Alt") == 0)   mod |= Mod1Mask; else 
      if ( strcmp( string, "Ctrl") == 0)  mod |= ControlMask; else 
      if ( strcmp( string, "Shift") == 0) mod |= ShiftMask; else 
      if (( strncmp( string, "Mod", 3) == 0) && ( strlen( string) == 4) && string[3] >= '1' && string[3] <= '5') {
         mod |= ( Mod1Mask << ( string[3] - '1'));
      } else {
         printf("Modificator '%s' cannot be resolved\n", string);
         exit(1);
      }
      string = d + 1;
   }
   k = XStringToKeysym( string);
   if ( k == NoSymbol) {
      printf("Symbol '%s' cannot be resolved\n", string);
      exit(1);
   }
   buttons[button][offset*2+0] = mod;
   buttons[button][offset*2+1] = XKeysymToKeycode(dpy, k);
   if ( debug)
      printf( "button %d, mods %d, keysym %d, keycode %d\n", 
         button, buttons[button][offset*2+0] , k, buttons[button][offset*2+1] );
}

int main(int argc, char *argv[])
{
   int i;
   if ( argc < 2) {
      printf("xmozmap: maps mouse buttons into key sequences. Useful for pointing devices with more than 3 buttons.\n\n");
      printf("format: xmozmap [-debug] [mouse_button=key[,key]] ... \n\n");
      printf("examples:\n");
      printf("  6=Alt+Left 7=Alt+Right\n");
      printf("  5=Shift+Ctrl+Escape,x,t,e,r,m,Enter 6=Shift+equals\n\n");
      printf("If you don't get what you've expected, try xev and see what keycodes are generated.\n\n");
      printf("author: Dmitry Karasik\n\n");
      exit(0);
   }
  
   if (!( dpy = XOpenDisplay(NULL))) {
      fprintf( stderr, "Can't open display, X may not be running.\n");
      exit(1);
   }


   memset( buttons, 0, sizeof(buttons));
   for (i = 1; i < argc; i++) {
      int num, j;
      char *d, *c;

      if ( strcmp( argv[i], "-debug") == 0) {
         debug = 1;
         continue;
      }
      
      c = strchr(( const char*) argv[i], '=');
      if ( !c) {
         fprintf( stderr, "argument '%s' is not valid: '=' symbol is missing\n", argv[i]);
         exit(1);
      }
      *c = 0;
      num = atoi( argv[i]);
      if ( num < 1 || num >= 256) {
         fprintf( stderr, "'%s' is not a valid mouse button identificator: must be within 1-256 range\n", argv[i]);
         exit(1);
      }
      c++;

      j = 2;
      d = c;
      while (( d = strchr( d, ','))) {
         j++;
         d++;
      }
      if ( !(buttons[num] = malloc( sizeof( unsigned int) * 2 * j))) {
         fprintf( stderr, "cannot allocate %d bytes\n", sizeof( unsigned int) * 2 * j);
         exit(1);
      }

      j = 0;
      while ( *c) {
         if (( d = strchr( c, ','))) *d = 0;
         parse( c, num, j++);
         if ( !d) break;
         c = d + 1;
      }
      buttons[num][j*2] = 0;
      buttons[num][j*2+1] = 0;
   }

  root=XDefaultRootWindow(dpy);
  for ( i = 0; i < 256; i++) {
     if ( buttons[i]) {
        XGrabButton(dpy,i,0,root,False,0,GrabModeAsync,GrabModeAsync,None,None);
        if (debug) printf( "Grabbed button %d\n", i);
     }
  }

  while(1) {
     XEvent ev;
     XNextEvent(dpy, &ev);
     if ( debug && ev. type==ButtonPress) 
        printf("ButtonPress event, button %d, mapping is %spresent\n", 
            ev. xbutton. button, buttons[ev. xbutton. button] ? "" : "not ");
     if( ev.type==ButtonPress && ev. xbutton. button < 256 && ev. xbutton. button >= 0 &&
         buttons[ev. xbutton. button]) {
        int foc_ret;
        Window foc;
        XGetInputFocus( dpy, &foc, &foc_ret);
        if ( foc != None) {
           XKeyEvent e;
           int * list = buttons[ev. xbutton. button];
           e. type = KeyPress;
           e. send_event = 1;
           e. display = dpy;
           e. window = foc;
           e. root = root;
           e. subwindow = foc;
           e. time = CurrentTime;
           e. same_screen = 1;
           i = 0;
           while ( list[i] != 0 || list[i+1] != 0) {
              e. state   = list[i];
              e. keycode = list[i+1];
              if ( debug) printf("%d %d\n", e. state, e.keycode);
              XSendEvent(dpy, foc, 1, 0xffffff, (XEvent*)&e);
              i += 2;
           }
        }
     }
  }
  for ( i = 0; i < 256; i++) {
     if ( buttons[i]) 
        XUngrabButton(dpy,i,0,root);
  }
  return 0;
}
